/*
 * plan9.c -- a graphics driver for Plan 9
 * (c) 2003 Andrey Mirtchovski
 */

/* #define PLAN9_DEBUG /* */

#ifdef PLAN9_DEBUG
#define MESSAGE print
#endif

#include <u.h>
#include <libc.h>
#include <draw.h>
#include <event.h>
#include <sys/time.h>
#include <errno.h>

#include "plan9.h"

static Image *colors[256];

enum{
	Font11 = 0,
	Font13,
	Font15,
	Font19,
	Font26,
	FontItal11,
	FontItal13,
	FontItal15,
	FontItal19,
	FontItal26,
	FontBold11,
	FontBold13,
	FontBold15,
	FontBold19,
	FontBold26,
	FontType11,
	FontType13,
	FontType15,
	FontType19,
	FontType26,
	FontLast,
};
Font *fnt[FontLast];	/* monospace and normal */

extern Gfxdrvr plan9_driver;
Gfxdev gd;			/* only a single window! */

char *buttons[] =
{
	"exit",
	0
};

Menu menu =
{
	buttons
};
Mouse m;

void
eresized(int new)
{
	if(new) {
		if(getwindow(display, Refnone) < 0)
			exits("can't reattach to window");
		gd.size.x2 = Dx(screen->r);
		gd.size.y2 = Dy(screen->r);
		gd.clip = gd.size;
		gd.resize_handler(&gd);
	}
}

static int	
plan9_translate_key(int key, int *k, int *f)
{
	*f = 0;
	if (key == 0177)		/* delete */
		exits(0);

	switch (key) {
	case '\b':	
		*k = KBD_BS;
		break;
	case '\t':	
		*k = KBD_TAB;
		break;
	case '\n':	
		*k = KBD_ENTER;
		break;
	case 13:	
		*k = KBD_END;
		break;
	case 033:	
		*k = KBD_ESC;
		break;
	case 0xf800:
		*k = KBD_DOWN;
		break;

	/* within Unicode private-use space */
	case 0xf00d:
		*k = KBD_HOME;
		break;
	case 0xf00e:
		*k = KBD_UP;
		break;
	case 0xf00f:
		*k = KBD_PAGE_UP;
		break;
	case 0xf011:
		*k = KBD_LEFT;
		break;
	case 0xf012:
		*k = KBD_RIGHT;
		break;
	case 0xf013:
		*k = KBD_PAGE_DOWN;
		break;
	case 0xf014:
		/* Ins, simulates '\f': refresh */
		*k = 'L';
		*f = KBD_CTRL;
		break;

	default:
		*k = key;
		break;
	}
	return 1;
}

static void
plan9_process_kbd(Event *ep)
{
	int key = ep->kbdc, k, f;

#ifdef PLAN9_DEBUG
MESSAGE("plan9_process_kbd: %c\n", key);
#endif
	if (plan9_translate_key(key, &k, &f))
		gd.keyboard_handler(&gd, k, f);
}

static Mouse oldm;

static void
plan9_process_mouse(Event *ep)
{
	Mouse m;
	int but = 0, evt = 0;

	m = ep->mouse;
	m.xy.x = m.xy.x - screen->r.min.x;
	m.xy.y = m.xy.y - screen->r.min.y;
#ifdef PLAN9_DEBUG
MESSAGE("plan9_process_mouse: %P, %d\n", m.xy, m.buttons);
#endif
	evt = B_MOVE;

	if(m.buttons & 1) {
		but = B_LEFT;
		if(oldm.buttons & 1)
			evt = B_DRAG;
		 else
			evt = B_DOWN;
	} else if(oldm.buttons & 1) {
		but = B_LEFT;
		evt = B_UP;
	}

	if(m.buttons & 2) {
		but = B_MIDDLE;
		if(oldm.buttons & 2)
			evt = B_DRAG;
		 else
			evt = B_DOWN;
	} else if(oldm.buttons & 2) {
		but = B_MIDDLE;
		evt = B_UP;
	}

	if(m.buttons & 4) {
		but = B_RIGHT;
		if(oldm.buttons & 4)
			evt = B_DRAG;
		 else
			evt = B_DOWN;
	} else if(oldm.buttons & 4) {
		but = B_RIGHT;
		evt = B_UP;
	}
	if(m.buttons & 8){
		gd.keyboard_handler(&gd, KBD_PAGE_UP, 0);
		return;
	}
	if(m.buttons & 16){
		gd.keyboard_handler(&gd, KBD_PAGE_DOWN, 0);
		return;
	}

	gd.mouse_handler(&gd, m.xy.x, m.xy.y, but|evt);

	oldm.buttons = m.buttons;
	oldm.msec = m.msec;
	oldm.xy = m.xy;
}

static void
redraw(void)
{
	Gfxdev *dev = &gd;
	struct rect r = {0, dev->size.x2, 0, dev->size.y2};

	t_redraw(dev, &r);
}

/*
 * closedisplay() and drawshutdown() in the event library set display to nil.
 * damn it to hell.
 */
static volatile int einited;
int evpipe[2];

static struct shared {
	ulong	key;
	Event	ev;
} sh;
static struct shared *shp = &sh;

extern int shootme;

static int
watchevents(void)
{
	char c;
	int pid;

	einit(Ekeyboard|Emouse);
	pid = rfork(RFPROC|RFMEM);
	if (pid == 0) {			/* child: event-watching process */
		for (;;) {
			if(display == nil)
				exits(0);
			shp->key = event(&shp->ev);
			/* there are events! */
			if (apewrite(evpipe[0], "x", 1) != 1)
				break;
			/* wait for event processing */
			if (aperead(evpipe[0], &c, 1) != 1)
				break;
		}
		shootme = 0;
		einited = 0;
		_exits(0);
	}
	/* parent */
	shootme = pid;
	seteventbit(evpipe[1]);
	einited = 1;
}

static int
processevent(Event *ep, ulong key)
{
	switch (key) {
	case Emouse:
		plan9_process_mouse(ep);
		break;
	case Ekeyboard:
		plan9_process_kbd(ep);
		break;
	default:
		fprint(2, "martian input event!\n");
		break;
	}
}

static int
processevents(void)
{
	char c;

	if (aperead(evpipe[1], &c, 1) != 1) {	/* wait for first event */
		einited = 0;
		return;
	}
	processevent(&shp->ev, shp->key);
	while (ecankbd() || ecanmouse()) {
		shp->key = eread(Ekeyboard|Emouse, &shp->ev);
		processevent(&shp->ev, shp->key);
	}
	if (apewrite(evpipe[1], "y", 1) != 1)	/* events are processed */
		einited = 0;
}

static void
update(void)
{
	if (display != nil)
		flushimage(display, 1);
}

static void
repaint(void)
{
	static time_t lastdraw, now;

	if (display != nil) {
		time(&now);
		if (now - lastdraw >= 5) {	/* catch stragglers */
			redraw();
			lastdraw = now;
		}
		update();
	}
}

int
plan9_select(int n, fd_set *readfds, fd_set *writefds, fd_set *exceptfds,
	struct timeval *timeout)
{
	int fds;
	static int first = 1;

#ifdef PLAN9_DEBUG
	MESSAGE("plan9_select\n");
#endif
	if (first) {
		first = 0;
		repaint();
	}
	check_bottom_halves();
	fds = select(n, readfds, writefds, exceptfds, timeout);
	if (fds > 0 && einited && isseteventbit(evpipe[1]))
		processevents();
	repaint();		/* enforce periodic screen update */
	return fds;
}

static Font *
plan9_loadfont(char *font)
{
	Font *f;

	f = openfont(display, font);
	if(f == nil) {
		print("%s: ", font);
		perror("cannot open font");
		exits("cannot load font");
	}

	return f;
}

static uchar *
plan9_init_driver(uchar *param, uchar *d)
{
	int i;
	char *tmp;

#ifdef PLAN9_DEBUG
	MESSAGE("plan9 init driver\n");
#endif
	if(initdraw(nil, nil, "flame") < 0)
		exits("initdraw failed");

	for (i = 0; i < 256; i++)
		colors[i] = allocimage(display, Rect(0, 0, 1, 1), screen->chan,
			1, cmap2rgba(i));

	/* normal fonts, the defaults are rather small */
	tmp = getenv("Font11");
	if(tmp != nil)
		fnt[Font11] = plan9_loadfont(tmp);
	else
		fnt[Font11] = plan9_loadfont("/lib/font/bit/lucidasans/unicode.6.font");

	tmp = getenv("Font13");
	if(tmp != nil)
		fnt[Font13] = plan9_loadfont(tmp);
	else
		fnt[Font13] = plan9_loadfont("/lib/font/bit/lucidasans/unicode.7.font");

	tmp = getenv("Font15");
	if(tmp != nil)
		fnt[Font15] = plan9_loadfont(tmp);
	else
		fnt[Font15] = plan9_loadfont("/lib/font/bit/lucidasans/unicode.8.font");

	tmp = getenv("Font19");
	if(tmp != nil)
		fnt[Font19] = plan9_loadfont(tmp);
	else
		fnt[Font19] = plan9_loadfont("/lib/font/bit/lucidasans/unicode.10.font");

	tmp = getenv("Font26");
	if(tmp != nil)
		fnt[Font26] = plan9_loadfont(tmp);
	else
		fnt[Font26] = plan9_loadfont("/lib/font/bit/lucidasans/unicode.13.font");

	/* italic fonts */
	tmp = getenv("FontItal11");
	if(tmp != nil)
		fnt[FontItal11] = plan9_loadfont(tmp);
	else
		fnt[FontItal11] = plan9_loadfont("/lib/font/bit/lucidasans/italicunicode.6.font");

	tmp = getenv("FontItal13");
	if(tmp != nil)
		fnt[FontItal13] = plan9_loadfont(tmp);
	else
		fnt[FontItal13] = plan9_loadfont("/lib/font/bit/lucidasans/italicunicode.7.font");

	tmp = getenv("FontItal15");
	if(tmp != nil)
		fnt[FontItal15] = plan9_loadfont(tmp);
	else
		fnt[FontItal15] = plan9_loadfont("/lib/font/bit/lucidasans/italicunicode.8.font");

	tmp = getenv("FontItal19");
	if(tmp != nil)
		fnt[FontItal19] = plan9_loadfont(tmp);
	else
		fnt[FontItal19] = plan9_loadfont("/lib/font/bit/lucidasans/italicunicode.10.font");

	tmp = getenv("FontItal26");
	if(tmp != nil)
		fnt[FontItal26] = plan9_loadfont(tmp);
	else
		fnt[FontItal26] = plan9_loadfont("/lib/font/bit/lucidasans/italicunicode.13.font");

	/* bold fonts */
	tmp = getenv("FontBold11");
	if(tmp != nil)
		fnt[FontBold11] = plan9_loadfont(tmp);
	else
		fnt[FontBold11] = plan9_loadfont("/lib/font/bit/lucidasans/boldunicode.6.font");

	tmp = getenv("FontBold13");
	if(tmp != nil)
		fnt[FontBold13] = plan9_loadfont(tmp);
	else
		fnt[FontBold13] = plan9_loadfont("/lib/font/bit/lucidasans/boldunicode.7.font");

	tmp = getenv("FontBold15");
	if(tmp != nil)
		fnt[FontBold15] = plan9_loadfont(tmp);
	else
		fnt[FontBold15] = plan9_loadfont("/lib/font/bit/lucidasans/boldunicode.8.font");

	tmp = getenv("FontBold19");
	if(tmp != nil)
		fnt[FontBold19] = plan9_loadfont(tmp);
	else
		fnt[FontBold19] = plan9_loadfont("/lib/font/bit/lucidasans/boldunicode.10.font");

	tmp = getenv("FontBold26");
	if(tmp != nil)
		fnt[FontBold26] = plan9_loadfont(tmp);
	else
		fnt[FontBold26] = plan9_loadfont("/lib/font/bit/lucidasans/boldunicode.13.font");

	/* typescript */
	tmp = getenv("FontType11");
	if(tmp != nil)
		fnt[FontType11] = plan9_loadfont(tmp);
	else
		fnt[FontType11] = plan9_loadfont("/lib/font/bit/lucidasans/typeunicode.6.font");

	tmp = getenv("FontType13");
	if(tmp != nil)
		fnt[FontType13] = plan9_loadfont(tmp);
	else
		fnt[FontType13] = plan9_loadfont("/lib/font/bit/lucidasans/typeunicode.7.font");

	tmp = getenv("FontType15");
	if(tmp != nil)
		fnt[FontType15] = plan9_loadfont(tmp);
	else
		fnt[FontType15] = plan9_loadfont("/lib/font/bit/lucidasans/typeunicode.9.font");

	tmp = getenv("FontType19");
	if(tmp != nil)
		fnt[FontType19] = plan9_loadfont(tmp);
	else
		fnt[FontType19] = plan9_loadfont("/lib/font/bit/lucidasans/typeunicode.12.font");

	tmp = getenv("FontType26");
	if(tmp != nil)
		fnt[FontType26] = plan9_loadfont(tmp);
	else
		fnt[FontType26] = plan9_loadfont("/lib/font/bit/lucidasans/typeunicode.16.font");

	eresized(0);
	watchevents();
	return nil;
}

static void
plan9_shutdown_driver(void)
{
#ifdef PLAN9_DEBUG
	MESSAGE("plan9_shutdown_driver\n");
#endif
	exits(0);
}

static Gfxdev*
plan9_init_device(void)
{
#ifdef PLAN9_DEBUG
	MESSAGE("plan9_init_device\n");
#endif
	gd.size.x1=0;
	gd.size.y1=0;
	gd.size.x2=Dx(screen->r);
	gd.size.y2=Dy(screen->r);

	gd.clip.x1=gd.size.x1;
	gd.clip.y1=gd.size.y1;
	gd.clip.x2=gd.size.x2;
	gd.clip.y2=gd.size.y2;
	gd.drv= &plan9_driver;
	gd.driver_data=nil;
	gd.user_data=0;

	return &gd;
}

/* close window */
static void
plan9_shutdown_device(Gfxdev *gd)
{
#ifdef PLAN9_DEBUG
	MESSAGE("plan9_shutdown_device\n");
#endif
	exits("shutdown device");
}

static int
plan9_get_filled_bitmap(Libitmap *bmp, long color)
{
#ifdef PLAN9_DEBUG
	MESSAGE("plan9_get_filled_bitmap\n");
#endif
	bmp->skip = bmp->x * 3;
	bmp->data = mallocz(bmp->skip * bmp->y, 0);
	bmp->flags = 0;
	return 0;
}

static int
plan9_get_empty_bitmap(Libitmap *bmp)
{
#ifdef PLAN9_DEBUG
	MESSAGE("plan9_get_empty_bitmap\n");
#endif
	bmp->skip=bmp->x*3;
	bmp->data=malloc(bmp->skip*bmp->y);
	bmp->flags=0;
	return 0;
}

static void
plan9_unregister_bitmap(Libitmap *bmp)
{
}

static void
plan9_fill_area(Gfxdev *dev, int x1, int y1, int x2, int y2, long color)
{
	Rectangle rt;
	int r, g, b;

	rt = rectaddpt(Rect(x1, y1, x2, y2), screen->r.min);
	b = (color>>16) & 0xff;
	g = (color>>8) & 0xff;
	r = (color) & 0xff;
#ifdef PLAN9_DEBUG
	MESSAGE("plan9_fill_area: %R; color: %ld\n", rt, color);
#endif
	draw(screen, rt, colors[rgb2cmap(r, g, b)], nil, ZP);
	// update();
}

static void
plan9_draw_hline(Gfxdev *dev, int left, int y, int right, long color)
{
	Point p1, p2;
	int r, g, b;

	b = (color>>16) & 0xff;
	g = (color>>8) & 0xff;
	r = (color) & 0xff;

	p1 = addpt(screen->r.min, Pt(left, y));
	p2 = addpt(screen->r.min, Pt(right, y));
#ifdef PLAN9_DEBUG
	MESSAGE("plan9_draw_hline: %R\n", Rpt(p1, p2));
#endif

	line(screen, p1, p2, Endsquare, Endsquare, 1, colors[rgb2cmap(r, g, b)], ZP);
	// update();

}

static void
plan9_draw_vline(Gfxdev *dev, int x, int top, int bottom, long color)
{
	Point p1, p2;
	int r, g, b;

	b = (color>>16) & 0xff;
	g = (color>>8) & 0xff;
	r = (color) & 0xff;

	p1 = addpt(screen->r.min, Pt(x, top));
	p2 = addpt(screen->r.min, Pt(x, bottom));

#ifdef PLAN9_DEBUG
	MESSAGE("plan9_draw_vline: %R\n", Rpt(p1, p2));
#endif

	line(screen, p1, p2, Endsquare, Endsquare, 1, colors[rgb2cmap(r, g, b)], ZP);
	// update();
}

static void
plan9_set_clip_area(Gfxdev *dev, struct rect *r)
{
	Rectangle cr;

	gd.clip = *r;
	cr = Rect(r->x1-1, r->y1-1, r->x2+1, r->y2+1);

#ifdef PLAN9_DEBUG
	print("plan9_set_clip_area: %R\n", cr);
#endif
	replclipr(screen, screen->repl, rectaddpt(cr, screen->r.min));
}

static void
draw_bitmap(Gfxdev *dev, Libitmap *bmp, int x, int y)
{
	Image *img;
	Rectangle r;
	ulong i;

	r = Rect(0, 0, bmp->x, bmp->y);
#ifdef PLAN9_DEBUG
	MESSAGE("plan9_draw_bitmap: size: %R at %P\n", r, Pt(x, y));
#endif
	if (display == nil)
		return;
	img = allocimage(display, r, RGB24, 0, 0);
	if(!img)
		return;

	loadimage(img, r, bmp->data, bmp->y*bmp->skip);
	draw(screen, rectaddpt(r, addpt(Pt(x, y), screen->r.min)), img, nil, ZP);
	freeimage(img);
}

static void
plan9_draw_bitmap(Gfxdev *dev, Libitmap *bmp, int x, int y)
{
	draw_bitmap(dev, bmp, x, y);
	update();
}

static void
plan9_draw_bitmaps(Gfxdev *dev, Libitmap **bmps, int n, int x, int y)
{
	int a;
#ifdef PLAN9_DEBUG
	MESSAGE("plan9_draw_bitmaps\n");
#endif
	print("plan9_draw_bitmaps\n");

	if (!bmps)
		return;
	for (a = 0; a < n; a++) {
		draw_bitmap(dev, bmps[a], x, y);
		x += bmps[a]->x;
	}
	update();
}

static void
plan9_register_bitmap(Libitmap *bmp)
{
}

int
plan9_hscroll(Gfxdev *dev, struct rect_set **set, int sc)
{
	Rectangle r;
	int i;

#ifdef PLAN9_DEBUG
	MESSAGE("hscroll\n");
#endif

	r = Rect(gd.clip.x1, gd.clip.y1, gd.clip.x2, gd.clip.y2);
	r = rectaddpt(r, screen->r.min);
	draw(screen, r, screen, nil, addpt(Pt(gd.clip.x1-sc, gd.clip.y1), screen->r.min));
	// update();
	return 1;
}


int plan9_vscroll(Gfxdev *dev, struct rect_set **set, int sc)
{
	Rectangle r;
	int i;

#ifdef PLAN9_DEBUG
	MESSAGE("vscroll\n");
#endif

	r = Rect(gd.clip.x1, gd.clip.y1, gd.clip.x2, gd.clip.y2);
	r = rectaddpt(r, screen->r.min);
	draw(screen, r, screen, nil, addpt(Pt(gd.clip.x1, gd.clip.y1-sc), screen->r.min));
	// update();
	return 1;
}


void *
plan9_prepare_strip(Libitmap *bmp, int top, int lines)
{
#ifdef PLAN9_DEBUG
	MESSAGE("plan9_prepare_strip\n");
#endif
	return (char *)bmp->data + bmp->skip*top;
}

void plan9_commit_strip(Libitmap *bmp, int top, int lines)
{
#ifdef PLAN9_DEBUG
	MESSAGE("plan9_commit_strip\n");
#endif
}

void
plan9_set_window_title(Gfxdev *gd, uchar *title)
{
	int fd;

#ifdef PLAN9_DEBUG
	MESSAGE("plan9_set_window_title: %s\n", title);
#endif
	if(title == nil)
		return;
	fd = open("/dev/label", OWRITE);
	if(fd >= 0) {
		write(fd, title, strlen((char *)title));
		close(fd);
	}
}

void
plan9_set_clipboard_text(Gfxdev *gd, uchar *text)
{
	int fd;

	if(text == nil)
		return;
#ifdef PLAN9_DEBUG
	MESSAGE("plan9_set_clipboard_text: %s\n", text);
#endif
	fd = open("/dev/snarf", OWRITE);
	if(fd >= 0) {
		write(fd, text, utflen((char *)text));
		close(fd);
	}
}

static uchar *
plan9_get_driver_param(void)
{
	return nil;
}

static int
plan9_exec(uchar *command, int fg)
{
#ifdef PLAN9_DEBUG
	MESSAGE("plan9_exec: %s\n", command);
#endif
	return 0;
}

static int
plan9_block(Gfxdev *)
{
#ifdef PLAN9_DEBUG
	MESSAGE("plan9_block\n");
#endif
	return 0;
}

static void
plan9_unblock(Gfxdev *)
{
#ifdef PLAN9_DEBUG
	MESSAGE("plan9_unblock\n");
#endif
}

static Font * 
getfont(struct style *style)
{
	Font *f;
	int height;

	if(style->mono_space != -1) {
		switch(style->height) {
		case 11: 	f = fnt[FontType11]; break;
		case 13: 	f = fnt[FontType13]; break;
		case 15: 	f = fnt[FontType15]; break;
		case 19: 	f = fnt[FontType19]; break;
		case 26: 	f = fnt[FontType26]; break;
		default: 	fprint(2, "martian mono font size: %d\n", style->height);
				f = fnt[FontType15]; break;
		}
	} else if(style->bold) {
		switch(style->height) {
		case 11: 	f = fnt[FontBold11]; break;
		case 13: 	f = fnt[FontBold13]; break;
		case 15: 	f = fnt[FontBold15]; break;
		case 19: 	f = fnt[FontBold19]; break;
		case 26: 	f = fnt[FontBold26]; break;
		default: 	fprint(2, "martian bold font size: %d\n", style->height);
				f = fnt[FontBold15]; break;
		}
	} else if(style->italic) {
		switch(style->height) {
		case 11: 	f = fnt[FontItal11]; break;
		case 13: 	f = fnt[FontItal13]; break;
		case 15: 	f = fnt[FontItal15]; break;
		case 19: 	f = fnt[FontItal19]; break;
		case 26: 	f = fnt[FontItal26]; break;
		default: fprint(2, "martian italic font size: %d\n", style->height);
				f = fnt[FontItal15]; break;
		}
	} else {
		switch(style->height) {
		case 11: 	f = fnt[Font11]; break;
		case 13: 	f = fnt[Font13]; break;
		case 15: 	f = fnt[Font15]; break;
		case 19: 	f = fnt[Font19]; break;
		case 26: 	f = fnt[Font26]; break;
		default: 	fprint(2, "martian normal font size: %d\n", style->height);
				f = fnt[Font15]; break;
		}
	}

	return f;
}

/* return the width of a character */
int 
plan9_charsize(struct style *style, char c)
{
	return stringnwidth(getfont(style), &c, 1);
}

/* return the length of a string */
int
plan9_stringsize(struct style *style, char *text)
{
	return stringwidth(getfont(style), text);
}

/* should return the width (in pixels) that the text occupies */
int 
plan9_print_text(struct graphics_driver *gd, struct
	graphics_device *device, int x, int y, struct style *style,
	char *text)

{
	Point sz,pos;
	Rectangle r;
	int xs;
	Font *f;
	char *c;

	while((c = strchr(text, 0x01)) != nil)
		*c = ' ';

	f = getfont(style);

	pos = addpt(Pt(x, y), screen->r.min);
	sz= stringsize(f, "W");
	sz.x*=strlen(text);
	r.min=pos;
	r.max=addpt(pos,sz);
	draw(screen, r, colors[rgb2cmap(style->r0, style->g0, style->b0)],
	    nil, ZP);

	stringbg(screen, pos, 
	    colors[rgb2cmap(style->r1, style->g1, style->b1)],
	    ZP, f, text,
	    colors[rgb2cmap(style->r0, style->g0, style->b0)],
	    ZP);
	update();
	sz=stringsize(f,text);
	return sz.x;
}

extern long (*color_pass_rgb)(int);

Gfxdrvr plan9_driver={
	(uchar *)"plan9",
	plan9_init_driver,
	plan9_init_device,
	plan9_shutdown_device,
	plan9_shutdown_driver,
	plan9_get_driver_param,
	plan9_get_empty_bitmap,
	plan9_get_filled_bitmap,
	plan9_register_bitmap,
	plan9_prepare_strip,
	plan9_commit_strip,
	plan9_unregister_bitmap,
	plan9_draw_bitmap,
	plan9_draw_bitmaps,
	color_pass_rgb,
	plan9_fill_area,
	plan9_draw_hline,
	plan9_draw_vline,
	plan9_hscroll,
	plan9_vscroll,
	plan9_set_clip_area,
	plan9_block,
	plan9_unblock,
	plan9_set_window_title,
	plan9_exec,
	195,			/* depth */
	0, 0,			/* size  */
	0,			/* flags */
	0,			/* codepage */
	(uchar *)"/bin/rc",	/* shell */
};
