typedef struct graphics_driver Gfxdrvr;
typedef struct graphics_device Gfxdev;
typedef struct bitmap Libitmap;

struct bitmap {
	int	x, y;	/* Dimensions */
	int	skip;	/* Bytes between vertically consecutive pixels */
	void	*data;	/* Pointer to room for topleft pixel */
	void	*user;	/* For implementing LRU algor by higher layer or what*/
	void	*flags;	/* Allocation flags for the driver */
};

struct rect {
	int	x1, x2, y1, y2;
};
struct graphics_device {
        /* Only graphics driver is allowed to write to this */

        struct rect size;	/* Size of the window */
        /* int left, right, top, bottom; */
	struct rect clip;
        /* right, bottom are coords of the first point outside clipping area */

        Gfxdrvr	*drv;
        void	*driver_data;

        /* Only user is allowed to write here, driver inits to zero's */
        void	*user_data;
        void	(*redraw_handler)(Gfxdev *dev, struct rect *r);
        void	(*resize_handler)(Gfxdev *dev);
        void	(*keyboard_handler)(Gfxdev *dev, int key, int flags);
        void	(*mouse_handler)(Gfxdev *dev, int x, int y, int buttons);
};

struct graphics_driver {
	uchar	*name;

	/* param is get from get_driver_param and saved into configure file */
	uchar *(*init_driver)(uchar *param, uchar *display);

	/* Creates new device and returns pointer to it */
	Gfxdev	*(* init_device)(void);
	void	(*shutdown_device)(Gfxdev *dev);  /* Destroys the device */
	void	(*shutdown_driver)(void);

	/* returns allocated string with parameter given to init_driver() */
	uchar	*(*get_driver_param)(void);

	/* dest must have x and y filled in when get_empty_bitmap is called */
	int	(*get_empty_bitmap)(Libitmap *dest);

	/* dest must have x & y filled in when get_filled_bitmap is called. */
	/* bitmap is created, pre-filled with n_bytes of pattern & registered */
	int	(*get_filled_bitmap)(Libitmap *dest, long color);
	void	(*register_bitmap)(Libitmap *bmp);
	void	*(*prepare_strip)(Libitmap *bmp, int top, int lines);
	void	(*commit_strip)(Libitmap *bmp, int top, int lines);

	/* Must not touch x and y. Suitable for re-registering. */
	void	(*unregister_bitmap)(Libitmap *bmp);
	void	(*draw_bitmap)(Gfxdev *dev, Libitmap *hndl, int x, int y);
	void	(*draw_bitmaps)(Gfxdev *dev, Libitmap **hndls, int n, int x, int y);

	/*
	 * rgb has gamma 1/display_gamma; 255 means exactly the largest
	 * sample the display is able to produce.
	 * Thus, if we have 3 bits for red, we will perform this code:
	 *	red = (red*7 +127)/255;
	 */
	long	(*get_color)(int rgb);
	void	(*fill_area)(Gfxdev *dev, int x1, int y1, int x2, int y2, long color);
	void	(*draw_hline)(Gfxdev *dev, int left, int y, int right, long color);
	void	(*draw_vline)(Gfxdev *dev, int x, int top, int bottom, long color);
	int	(*hscroll)(Gfxdev *dev, struct rect_set **set, int sc);
	int	(*vscroll)(Gfxdev *dev, struct rect_set **set, int sc);

	/*
	 * When scrolling, empty spaces will have undefined contents. returns:
	 *  0 - the caller should not care about redrawing, redraw will be sent
	 *  1 - the caller should redraw uncovered area.
	 * when set is not NULL, rectangles in the set (uncovered area)
	 * should be redrawn.
	 */
	void	(*set_clip_area)(Gfxdev *dev, struct rect *r);

	/* restore old videomode and disable drawing on terminal */
	int	(*block)(Gfxdev *dev);
	void	(*unblock)(Gfxdev *dev);	/* reenable the terminal */
	/*
	 * set window title. title is in utf-8 encoding -- you should recode
	 * it to device charset. if device doesn't support titles (svgalib,
	 * framebuffer), this should be NULL, not empty function!
	 */
	void	(*set_title)(Gfxdev *dev, uchar *title);

	/*
	 * -if !NULL executes command on this graphics device,
	 * -if NULL links uses generic (console) command executing functions
	 * -return value is the same as of the 'system' syscall
	 * -if flag is !0, run command in separate shell
	 * else run command directly
	 */
	int	(*exec)(uchar *command, int flag);

	/*
	 * Data layout
	 * depth
	 *  8 7 6 5 4 3 2 1 0
	 * +-+---------+-----+
	 * | |         |     |
	 * +-+---------+-----+
	 *
	 * 0 - 2 Number of bytes per pixel in passed bitmaps
	 * 3 - 7 Number of significant bits per pixel -- 1, 4, 8, 15, 16, 24
	 * 8   0-- normal order, 1-- misordered.  Has the same value as
	 *	vga_misordered from the VGA mode.
	 *
	 * This number is to be used by the layer that generates images.
	 * Memory layout for 1 bytes per pixel is:
	 * 2 colors:
	 *  7 6 5 4 3 2 1 0
	 * +-------------+-+
	 * |      0      |B| B is The Bit. 0 black, 1 white
	 * +-------------+-+
	 *
	 * 16 colors:
	 *  7 6 5 4 3 2 1 0
	 * +-------+-------+
	 * |   0   | PIXEL | Pixel is 4-bit index into palette
	 * +-------+-------+
	 *
	 * 256 colors:
	 *  7 6 5 4 3 2 1 0
	 * +---------------+
	 * |  --PIXEL--    | Pixels is 8-bit index into palette
	 * +---------------+
	 */
	int	depth;

	/* size of screen. only for drivers that use virtual devices */
	int	x, y;
	int	flags;			/* GD_xxx flags */
	int	codepage;
	/*
	 * -if exec is NULL string is unused
	 * -otherwise this string describes shell to be executed by the
	 * exec function, the '%' char means string to be executed
	 * -shell cannot be NULL; if exec is !NULL and shell is empty,
	 * exec should use some default shell (e.g. "xterm -e %")
	 */
	uchar	*shell;
};

struct font {
	uchar *family;
	uchar *weight;
	uchar *slant;
	uchar *adstyl;
	uchar *spacing;
	int begin; /* Begin in the letter stream */
	int length; /* Length in the letter stream */
};

struct style{
	int refcount;
	unsigned char r0, g0, b0, r1, g1, b1;
       	/* ?0 are background, ?1 foreground.
	 * These are unrounded 8-bit sRGB space 
	 */
	int height;
	int flags; /* non-zero means underline */
	long underline_color; /* Valid only if flags are nonzero */
	int *table; /* First is refcount, then n_fonts entries. Total
                     * size is n_fonts+1 integers.
		     */
	int mono_space; /* -1 if the font is not monospaced
			 * width of the space otherwise
			 */
	int mono_height; 
	int bold;
	int italic;
};

#define KBD_SHIFT	1
#define KBD_CTRL	2
#define KBD_ALT		4

#define KBD_ENTER	-0x100
#define KBD_BS		-0x101
#define KBD_TAB		-0x102
#define KBD_ESC		-0x103
#define KBD_LEFT	-0x104
#define KBD_RIGHT	-0x105
#define KBD_UP		-0x106
#define KBD_DOWN	-0x107
#define KBD_INS		-0x108
#define KBD_DEL		-0x109
#define KBD_HOME	-0x10a
#define KBD_END		-0x10b
#define KBD_PAGE_UP	-0x10c
#define KBD_PAGE_DOWN	-0x10d

#define BM_BUTT		15
#define B_LEFT		0
#define B_MIDDLE	1
#define B_RIGHT		2

#define BM_ACT		48
#define B_DOWN		0
#define B_UP		16
#define B_DRAG		32
#define B_MOVE		48
